#include <be/AppKit.h>
#include <be/StorageKit.h>
#include <be/interface/Alert.h>
#include <be/support/Autolock.h>
#include <map>

class NodeWin;

class NodeApp: public BApplication {
	bool win;
	BPath mPath;
	BEntry mEntry;
	BDirectory mDir;
	
	struct NodeCompare {
		bool operator()(const node_ref& n1, const node_ref& n2) const
		{return n1.device<n2.device || (n1.device==n2.device && n1.node<n2.node);}
	};
	typedef map< node_ref,BPath,NodeCompare,allocator< pair<node_ref,BPath> > > map_type;
	map_type node;
	
  public:
	NodeApp(bool w): BApplication("application/x-vnd.Gargoyle-NodeMonitor"),win(w)
	{	watch_node(NULL,B_WATCH_MOUNT,be_app_messenger);	}
	~NodeApp(){}
	
	virtual void ReadyToRun();
	virtual void ArgvReceived(int32, char **);
	virtual void RefsReceived(BMessage *);
	virtual void MessageReceived(BMessage *);
	
	status_t Monitor(const entry_ref& e) {
		node_ref n;
		
		mEntry.SetTo(&e);
		mEntry.GetNodeRef(&n);
		mEntry.GetPath(&mPath);
		printf("monitoring '%s'\n",mPath.Path());
		node[n]=mPath;
		
		return watch_node(&n, B_WATCH_ALL|B_WATCH_DIRECTORY, be_app_messenger);
	}
};


class NodeWin: public BWindow {
	NodeApp* app;
	entry_ref ref;

  public:
	NodeWin(): BWindow(BRect(100,100,200,200),"NodeMonitor",B_TITLED_WINDOW,0),
		app((NodeApp*)be_app)
	{	Show();	}
	
	virtual void MessageReceived(BMessage *msg){
		if (msg->what!=B_SIMPLE_DATA) BWindow::MessageReceived(msg);
		BAutolock l(app);
		for (int i=0;msg->FindRef("refs",i,&ref)==B_OK;i++)
			app->Monitor(ref);
	}
	virtual bool QuitRequested()
	{
		int32 index=(new BAlert(	"Quit NodeMonitor?", 
								"Quit or just close the window?",
								"Cancel", "Close", "Quit"		))->Go();
		if (index>1)
			app->PostMessage(B_QUIT_REQUESTED);
		return index>0;
	}
};

void main(int argc, char **)
{
	NodeApp app(argc==1);
	app.Run();
}

void NodeApp::ReadyToRun()
{
	if (win) new NodeWin();
}
void NodeApp::ArgvReceived(int32 argc, char **argv)
{
	entry_ref ref;
	while (--argc>0)
		if (get_ref_for_path(argv[argc],&ref)==B_OK) Monitor(ref);
}
void NodeApp::RefsReceived(BMessage *msg)
{
	entry_ref ref;
	for (int i=0; msg->FindRef("refs",i,&ref)==B_OK; i++)
		Monitor(ref);
}
void NodeApp::MessageReceived(BMessage *message)
{
	node_ref n;
	entry_ref e;
	map_type::iterator i;
	
	if (message->what==B_NODE_MONITOR)
	switch (message->FindInt32("opcode")){
	case B_ENTRY_CREATED:
		n.device=message->FindInt32("device");
		message->FindInt64("directory",&n.node);
		mDir.SetTo(&n);
		mDir.GetEntry(&mEntry);
		mEntry.GetPath(&mPath);
		printf("'%s/%s' created\n",
			mPath.Path(),
			message->FindString("name"));
		break;
	
	case B_ENTRY_REMOVED:
		n.device=message->FindInt32("device");
		message->FindInt64("directory",&n.node);
		mDir.SetTo(&n);
		mDir.GetEntry(&mEntry);
		mEntry.GetPath(&mPath);
		
		message->FindInt64("node",&n.node);
		i=node.find(n);
		if (i!=node.end()) mPath.Append("(unknown)");
		printf("'%s' removed",	i!=node.end()?(*i).second.Path():mPath.Path());
		node.erase(n);
		break;
	
	case B_ENTRY_MOVED:
		e.device=message->FindInt32("device");
		message->FindInt64("to directory",&e.directory);
		e.set_name(message->FindString("name"));
		mEntry.SetTo(&e);
		mEntry.GetPath(&mPath);
		mEntry.GetNodeRef(&n);
		
		i=node.find(n);
		if (i!=node.end()) {
			printf("'%s' -> '%s'\n",(*i).second.Path(), mPath.Path());
			(*i).second=mPath;
		} else { // We were watching the directory
			n.node=e.directory;
			i=node.find(n);
			
			if (i!=node.end()){
				message->FindInt64("from directory",&n.node);
				mDir.SetTo(&n);
				mDir.GetEntry(&mEntry);
				mEntry.GetPath(&mPath);
				
				printf("'%s/%s' -> '%s/%s'\n",mPath.Path(),e.name,(*i).second.Path(),e.name);
			} else {// must be "from directory" that we're watching
				mDir.SetTo(&n);
				mDir.GetEntry(&mEntry);
				mEntry.GetPath(&mPath);
				message->FindInt64("from directory",&n.node);
				i=node.find(n);
				
				printf("'%s/%s' -> '%s/%s'\n",(*i).second.Path(),e.name,mPath.Path(),e.name);
			}
		}	
		break;
	
	case B_STAT_CHANGED:
		n.device=message->FindInt32("device");
		message->FindInt64("node",&n.node);
		i=node.find(n);
		printf("'%s' stat changed\n",
			i!=node.end()?(*i).second.Path():"(unknown)");
		break;
	
	case B_ATTR_CHANGED:
		n.device=message->FindInt32("device");
		message->FindInt64("node",&n.node);
		i=node.find(n);
		printf("'%s' attribute changed\n",
			i!=node.end()?(*i).second.Path():"(unknown)");
		break;
	
	case B_DEVICE_MOUNTED:
		n.device=message->FindInt32("device");
		message->FindInt64("directory",&n.node);
		mDir.SetTo(&n);
		mDir.GetEntry(&mEntry);
		mEntry.GetPath(&mPath);
		printf("'%s' (device %i) mounted\n",
			mPath.Path(),
			message->FindInt32("new device"));
		n.node=0;
		node[n]=mPath;
		break;
	
	case B_DEVICE_UNMOUNTED:
		n.device=message->FindInt32("device");
		n.node=0;
		i=node.find(n);
		printf("'%s' (device %i) unmounted\n",
			i!=node.end()?(*i).second.Path():"(unknown)",n.device);
		for (i=node.begin();i!=node.end();i++)
			if ((*i).first.device==n.device) node.erase(i);
		
		break;
	}
	else MessageReceived(message);
}